/**
 * TrueReality - N64/exceptions.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>

#include <config.h>
#include <parser_extern.h>

#include "registers.h"
#include "exceptions.h"

#ifdef DEBUG
        #include <debug.h>
#endif
   




void rs4300i_trap_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Tr;

#ifdef DEBUG
        if(prefs.debug & DBG_EXC_TR)
                printf("Exception (Tr) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_trap_exception() */





void rs4300i_break_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Bp;

#ifdef DEBUG
        if(prefs.debug & DBG_EXC_BP)
                printf("Exception (Bp) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_break_exception() */





void rs4300i_syscall_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Sys;

#ifdef DEBUG
        if(prefs.debug & DBG_EXC_SYS)
                printf("Exception (Sys) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_syscall_exception() */





void rs4300i_integer_overflow_exception()
{
        
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Ov;

#ifdef DEBUG
        if(prefs.debug & DBG_EXC_OV)
                printf("Exception (Ov) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_integer_overflow_exception() */





void rs4300i_floating_point_exception()
{
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_FPE;

#ifdef DEBUG
        if(prefs.debug & DBG_EXC_FPE)
                printf("Exception (FPE) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_integer_overflow_exception() */





void rs4300i_reserved_instruction_exception()
{
        reg.cpr[0][CAUSE] &= ~EXC_CODE__MASK;
        reg.cpr[0][CAUSE] |= EXC_CODE_Ri;

#ifdef DEBUG
        if(prefs.debug & DBG_EXC_RI)
                printf("Exception (Ri) @ %16llx\n", reg.cpr[0][COUNT]); fflush(stdout);
#endif



        jump_to_common_exception_vector();

} /* void rs4300i_reserved_instruction_exception() */










void jump_to_common_exception_vector()
{
        reg.cpr[0][STATUS] |= 0x00000002;               /* set EXL in the STATUS reg        */
        reg.cpr[0][EPC]  = reg.pc;                      /* store PC in EPC                  */

        if(reg.delay == EXEC_DELAY)
        {
                reg.cpr[0][CAUSE] |= 0x80000000;                /* set BD bit in CAUSE reg          */
                reg.cpr[0][EPC]   -= 4;                         /* store pc before branch instr     */

        } /* if(reg.delay == EXEC_DELAY) */
        else
        {
                reg.cpr[0][CAUSE] &= ~0x80000000;               /* clr BD bit in CAUSE reg          */
        }

        reg.pc = 0x80000180;                            /* set pc to 0x80000180 (common exception vector) */
        reg.delay = NO_DELAY;                           /* next instruction must NOT be the instr in the delay slot !!! */

} /* void jump_to_common_exception_vector() */

