/**
 * TrueReality - N64/registers.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/* every 'export' in this file resides in 'cpu.c' */
/* this file could also be named as 'cpu_export.h' */





#define __REGISTERS_H


#ifndef __TYPE_SIZES_H
        #include "type_sizes.h"
#endif





struct rs4300i_tlb
{
        WORD    hh;
        WORD    hl;
        WORD    lh;
        WORD    ll;
        BIT     g;
};





/* all "real" values are in cpu.c */

/* rs4300i registers */

struct rs4300i_reg
{
        sDWORD  r[32];      /* General Purpose Registers     (GPR) */
        sDWORD  cpr[3][32]; /* CoProcessor Registers         (CPR) */
        sDWORD  ccr[3][32]; /* Coprocessor Control Registers (CCR) */

        DWORD   hi, lo;     /* multiply and divide registers */
        
        WORD    pc;         /* program counter */



        /* add emulation reg */
        
        WORD    pc_delay;   /* delayed program counter */
        int     delay;      /* this is for correct emulation of the 6 stage pipeline */
                            /* if(delay == NO_DELAY)   { next instr is pc + 4 } */
                            /* if(delay == DO_DELAY)   { next instr is the delayed instruction } */
                            /* if(delay == EXEC_DELAY) { exec instr @ delayed pc - set in switch} */
        
        int     do_or_check_sthg;
                            /* if '!0' then the rs4300i has to do or check sthg */
                            /* look at the defines that follow this struct */

        int     halt;       /* if '!0' then stop execution */
        WORD    code;       /* current code (needed for speed) */
        
        BIT     llbit;

        struct rs4300i_tlb tlb[32];



}; /* struct rs4300i_reg */



/* rs4300i_reg.delay */

#define NO_DELAY        0
#define DO_DELAY        1
#define EXEC_DELAY      2





/* defines for 'rs4300i_reg.do_or_check_sthg' */

#define RS4300I_DO_OR_CHECK_NOTHING             0x0000
#define RS4300I_CHECK_CPU_INTERRUPTS            0x0001
#define RS4300I_CHECK_MI_INTERRUPTS             0x0002
#define RS4300I_DO_SI_INTERRUPT                 0x0004
#define RS4300I_DO_SP_INTERRUPT                 0x0008
#define RS4300I_DO_DP_INTERRUPT                 0x0010
#define RS4300I_DO_AI_INTERRUPT                 0x0020
#define RS4300I_DO_PI_INTERRUPT                 0x0040
#define RS4300I_DO_INTEGER_OVERFLOW_EXCEPTION   0x0080
#define RS4300I_DO_FLOATING_POINT_EXCEPTION     0x0100
#define RS4300I_DO_SYSCALL_EXCEPTION            0x0200
#define RS4300I_DO_TRAP_EXCEPTION               0x0400
#define RS4300I_DO_BREAK_EXCEPTION              0x0800
#define RS4300I_DO_RESERVED_INSTRUCTION_EXCEPTION       0x1000





extern struct rs4300i_reg reg;





#define INDEX           0
#define RANDOM          1
#define ENTRYLO0        2
#define ENTRYLO1        3
#define CONTEXT         4
#define PAGEMASK        5
#define WIRED           6
#define BADVADDR        8
#define COUNT           9
#define ENTRYHI         10
#define COMPARE         11
#define STATUS          12
#define CAUSE           13
#define EPC             14
#define PRID            15
#define CONFIG          16
#define LLADDR          17
#define WATCHLO         18
#define WATCHHI         19
#define XCONTEXT        20
#define ECC             26
#define CACHEERR        27
#define TAGLO           28
#define TAGHI           29
#define ERROREPC        30



/******************************************************************************\
*                                                                              *
* cop0 registers                                                               *
*                                                                              *
* cpr[0][00] = Index                                                           *
* cpr[0][01] = Random                                                          *
* cpr[0][02] = EntryLo0                                                        *
* cpr[0][03] = EntryLo1                                                        *
* cpr[0][04] = Context                                                         *
* cpr[0][05] = PageMask                                                        *
* cpr[0][06] = Wired                                                           *
* cpr[0][07] =                                                                 *
* cpr[0][08] = BadVAddr                                                        *
* cpr[0][09] = Count (should be 32bit - i use 64bit because of interrupts)     *
* cpr[0][10] = EntryHi                                                         *
* cpr[0][11] = Compare                                                         *
* cpr[0][12] = Status                                                          *
* cpr[0][13] = Cause                                                           *
* cpr[0][14] = EPC                                                             *
* cpr[0][15] = PRId                                                            *
* cpr[0][16] = Config                                                          *
* cpr[0][17] = LLAddr                                                          *
* cpr[0][18] = WatchLo                                                         *
* cpr[0][19] = WatchHi                                                         *
* cpr[0][20] = XContext                                                        *
* cpr[0][21] =                                                                 *
* cpr[0][22] =                                                                 *
* cpr[0][23] =                                                                 *
* cpr[0][24] =                                                                 *
* cpr[0][25] =                                                                 *
* cpr[0][26] = ECC                                                             *
* cpr[0][27] = CacheErr                                                        *
* cpr[0][28] = TagLo                                                           *
* cpr[0][29] = TagHi                                                           *
* cpr[0][30] = ErrorEPC                                                        *
* cpr[0][31] =                                                                 *
*                                                                              *
\******************************************************************************/

