/**
 * TrueReality - Display/local/amigados/amigados.c
 * Copyright (C) 1998, 1999 Mathias "AmiDog" Roslund
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/graphics.h>
#include <powerup/ppcproto/dos.h>
#include <math.h>

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>

#include "keyboard.h"

#ifdef JOYSTICK
  #include "../joystick/joystick.h"
#endif

#ifdef DEBUG
  #include "debug.h"
#endif

extern struct Window *window;

WORD CheckVisual()
{
  static HWORD keystate = 0;
  static sBYTE stick_x = 0;
  static sBYTE stick_y = 0;
  struct IntuiMessage *imsg = NULL;
  ULONG imCode,imClass;
  int x, y;

     /* Check for IDCMP messages */
     if((imsg=(struct IntuiMessage *)GetMsg(window->UserPort)))
     {
        imClass=imsg->Class;
        imCode=imsg->Code;
        x=imsg->MouseX;
        y=imsg->MouseY;
        ReplyMsg((struct Message *)imsg);

#ifdef JOYSTICK
        if(prefs.joystick == USE_JOYSTICK)
          ReadJoystick(&stick_x, &stick_y, &a, &b);
#endif

        /* FocusChange */

        if(imClass==IDCMP_ACTIVEWINDOW)
        {
          puts("   ***   F O C U S   I N   ***");
        }
        else if (imClass==IDCMP_INACTIVEWINDOW)
        {
          puts("   ***   F O C U S   O U T   ***");
        }

        /* KeyPress and KeyReleases */

        if(imClass==IDCMP_RAWKEY)
        {
                if(imCode<128) /* Key is being pressed */
                {
                        switch(imCode)
                        {
                            case 69: /* ESC */
                                dispatch(SHUTDOWN, 0);

                            case 87: /* F8, Refresh Visual */
                                dispatch(DISPLAY_REFRESH, 0);
                                break;

                            case 16: /* Q, Start Button */
                                keystate |= 0x1000;
                                break;

                            case 17: /* W, A-Button */
                                keystate |= 0x8000;
                                break;

                            case 18: /* E, B-Button */
                                keystate |= 0x4000;
                                break;

                            case 62: /* 8 (num), C-Button UP */
                                keystate |= 0x0008;
                                break;

                            case 30: /* 2 (num), C-Button DOWN */
                                keystate |= 0x0004;
                                break;

                            case 45: /* 4 (num), C-Button LEFT */
                                keystate |= 0x0002;
                                break;

                            case 47: /* 6 (num), C-Button RIGHT */
                                keystate |= 0x0001;
                                break;

                            case 76: /* UP (arrow), D-Button UP */
                                keystate |= 0x0800;
                                break;

                            case 77: /* DOWN (arrow), D-Button DOWN */
                                keystate |= 0x0400;
                                break;

                            case 79: /* LEFT (arrow), D-Button LEFT */
                                keystate |= 0x0200;
                                break;

                            case 78: /* RIGTH (arrow), D-Button RIGHT */
                                keystate |= 0x0100;
                                break;

                            case 32: /* A, R-Button */
                                keystate |= 0x0010;
                                break;

                            case 33: /* S, L-Button */
                                keystate |= 0x0020;
                                break;

                            case 34: /* D, Z-Button */
                                keystate |= 0x2000;
                                break;

                            case 39: /* K, Stick Up */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                if(stick_y < 120)
                                        stick_y += 0x08;
                                break;

                            case 38: /* J, Stick Down */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                if(stick_y > -120)
                                        stick_y -= 0x08;
                                break;

                            case 40: /* L, Stick Right */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                if(stick_x < 120)
                                        stick_x += 0x08;
                                break;

                            case 37: /* H, Stick Left */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                if(stick_x > -120)
                                        stick_x -= 0x08;
                                break;

                        } /* switch(imCode) */
                }
                else /* (imCode>127), Key released */
                {
                imCode-=128; /* Remove "Key released" bit */
                        switch(imCode)
                        {
                            case 16: /* Q, Start Button */
                                keystate &= ~0x1000;
                                break;

                            case 17: /* W, A-Button */
                                keystate &= ~0x8000;
                                break;

                            case 18: /* E, B-Button */
                                keystate &= ~0x4000;
                                break;

                            case 62: /* 8 (num), C-Button UP */
                                keystate &= ~0x0008;
                                break;

                            case 30: /* 2 (num), C-Button DOWN */
                                keystate &= ~0x0004;
                                break;

                            case 45: /* 4 (num), C-Button LEFT */
                                keystate &= ~0x0002;
                                break;

                            case 47: /* 6 (num), C-Button RIGHT */
                                keystate &= ~0x0001;
                                break;

                            case 76: /* UP (arrow), D-Button UP */
                                keystate &= ~0x0800;
                                break;

                            case 77: /* DOWN (arrow), D-Button DOWN */
                                keystate &= ~0x0400;
                                break;

                            case 79: /* LEFT (arrow), D-Button LEFT */
                                keystate &= ~0x0200;
                                break;

                            case 78: /* RIGHT (arrow), D-Button RIGHT */
                                keystate &= ~0x0100;
                                break;

                            case 32: /* A, R-Button */
                                keystate &= ~0x0010;
                                break;

                            case 33: /* S, L-Button */
                                keystate &= ~0x0020;
                                break;

                            case 34: /* D, Z-Button */
                                keystate &= ~0x2000;
                                break;

                        } /* switch(imCode) */

                } /* (imCode>127), Key released */

        } /* imClass=IDCMP_RAWKEY */

     } /* Check for IDCMP messages */

#ifdef DEBUG_CONTROLLER
        if(prefs.debug & DBG_CONTR_BUTTONS)
                printf
                (
                        "Controller: "
                        "0x%04hx %+4d %+4d (0x%08lx) (0xBBBBXXYY) "
                        "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",

                        keystate, stick_x, stick_y,
                        (WORD)(keystate << 16) | ((WORD)(HWORD)(stick_x << 8)) | ((WORD)(BYTE)stick_y),
                        (keystate & 0x8000) ? "A "      : "",
                        (keystate & 0x4000) ? "B "      : "",
                        (keystate & 0x2000) ? "Z "      : "",
                        (keystate & 0x1000) ? "START "  : "",
                        (keystate & 0x0800) ? "Dup "    : "",
                        (keystate & 0x0400) ? "Ddown "  : "",
                        (keystate & 0x0200) ? "Dleft "  : "",
                        (keystate & 0x0100) ? "Dright " : "",
                        (keystate & 0x0080) ? ""        : "",
                        (keystate & 0x0040) ? ""        : "",
                        (keystate & 0x0020) ? "L "      : "",
                        (keystate & 0x0010) ? "R "      : "",
                        (keystate & 0x0008) ? "Cup "    : "",
                        (keystate & 0x0004) ? "Cdown "  : "",
                        (keystate & 0x0002) ? "Cleft "  : "",
                        (keystate & 0x0001) ? "Cright " : "" 
                );
#endif

  return(((WORD)(keystate << 16)) | ((WORD)(HWORD)(stick_x << 8)) | ((WORD)(BYTE)stick_y));

} /* WORD CheckVisual() */

