/**
 * TrueReality - input/keyboard/x11.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>
#include <x11_extern.h>

#include "keyboard.h"

#ifdef JOYSTICK
        #include "../joystick/joystick.h"
#endif
#ifdef DEBUG
        #include <debug.h>
#endif

XKeyboardState     xkeyboardstate;








WORD CheckVisual()
{
        XEvent          event;
        KeySym          key;

        static HWORD    keystate = 0;

#ifdef JOYSTICK
        int             a, b;
#endif
        static sBYTE    stick_x = 0;
        static sBYTE    stick_y = 0;





#ifdef JOYSTICK
        if(prefs.joystick == USE_JOYSTICK)
                ReadJoystick(&stick_x, &stick_y, &a, &b);
#endif


        /* FocusChange */

        if(XCheckWindowEvent(dsp, win, FocusChangeMask, &event))
        {
                if(event.type == FocusIn)
                {
                        puts("   ***   F O C U S   I N   ***");
                        
                        XGetKeyboardControl(dsp, &xkeyboardstate);

                        if(xkeyboardstate.global_auto_repeat == AutoRepeatModeOn)
                                XAutoRepeatOff(dsp);
                }
                else
                {
                        puts("   ***   F O C U S   O U T   ***");

                        if(xkeyboardstate.global_auto_repeat == AutoRepeatModeOn)
                                XAutoRepeatOn(dsp);
                }
        }



        /* ButtonPress / ButtonRelease */ 

        if(XCheckWindowEvent(dsp, win, ButtonPressMask | ButtonReleaseMask, &event))
        {
        }



        /* KeyPress and KeyReleases */

        if(XCheckWindowEvent(dsp, win, KeyPressMask | KeyReleaseMask, &event))
        {
                key = XLookupKeysym((XKeyEvent *)&event, 0);

                if(event.type == KeyPress)
                {
                        switch(key)
                        {
                            case XK_Escape:
                                dispatch(SHUTDOWN, 0);

                            case XK_F8:
                                dispatch(DISPLAY_REFRESH, 0);
                                break;

                            case XK_q:           /* Start Button */
                                keystate |= 0x1000;
                                break;

                            case XK_w:                  /* A-Button */
                                keystate |= 0x8000;
                                break;

                            case XK_e:                  /* B-Button */
                                keystate |= 0x4000;
                                break;

                            case XK_Home:               /* C-Button UP */
                                keystate |= 0x0008;
                                break;

                            case XK_End:                /* C-Button DOWN */
                                keystate |= 0x0004;
                                break;

                            case XK_Delete:          /* C-Button LEFT */
                                keystate |= 0x0002;
                                break;

                            case XK_Next:          /* C-Button RIGHT */
                                keystate |= 0x0001;
                                break;

                            case XK_Up:              /* D-Button UP */
                                keystate |= 0x0800;
                                break;

                            case XK_Down:            /* D-Button DOWN */
                                keystate |= 0x0400;
                                break;

                            case XK_Left:            /* D-Button LEFT */
                                keystate |= 0x0200;
                                break;

                            case XK_Right:           /* D-Button RIGHT */
                                keystate |= 0x0100;
                                break;

                            case XK_a:          /* R-Button */
                                keystate |= 0x0010;
                                break;

                            case XK_s:            /* L-Button */
                                keystate |= 0x0020;
                                break;

                            case XK_d:                  /* Z-Button */
                                keystate |= 0x2000;
                                break;

                            case XK_k:   /* Stick Up */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                        if(stick_y < 120)
                                                stick_y += 0x08;
                                break;

                            case XK_j:   /* Stick Down */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                        if(stick_y > -120)
                                                stick_y -= 0x08;
                                break;

                            case XK_l:   /* Stick Right */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                        if(stick_x < 120)
                                                stick_x += 0x08;
                                break;

                            case XK_h:   /* Stick Left */
#ifdef JOYSTICK
                                if(prefs.joystick != USE_JOYSTICK)
#endif
                                        if(stick_x > -120)
                                                stick_x -= 0x08;
                                break;

                        } /* switch(key) */
                }
                else /* (event.type == KeyRelease) */
                {
                        switch(key)
                        {
                            case XK_q:           /* Start Button */
                                keystate &= ~0x1000;
                                break;

                            case XK_w:                  /* A-Button */
                                keystate &= ~0x8000;
                                break;

                            case XK_e:                  /* B-Button */
                                keystate &= ~0x4000;
                                break;

                            case XK_Home:               /* C-Button UP */
                                keystate &= ~0x0008;
                                break;

                            case XK_End:                /* C-Button DOWN */
                                keystate &= ~0x0004;
                                break;

                            case XK_Delete:          /* C-Button LEFT */
                                keystate &= ~0x0002;
                                break;

                            case XK_Next:          /* C-Button RIGHT */
                                keystate &= ~0x0001;
                                break;

                            case XK_Up:              /* D-Button UP */
                                keystate &= ~0x0800;
                                break;

                            case XK_Down:            /* D-Button DOWN */
                                keystate &= ~0x0400;
                                break;

                            case XK_Left:            /* D-Button LEFT */
                                keystate &= ~0x0200;
                                break;

                            case XK_Right:           /* D-Button RIGHT */
                                keystate &= ~0x0100;
                                break;

                            case XK_a:          /* R-Button */
                                keystate &= ~0x0010;
                                break;

                            case XK_s:            /* L-Button */
                                keystate &= ~0x0020;
                                break;

                            case XK_d:                  /* Z-Button */
                                keystate &= ~0x2000;
                                break;

                        } /* switch(key) */

                } /* (event.type == KeyRelease) */

        } /* if(XCheckWindowEvent(dsp, win, KeyPressMask | KeyReleaseMask, &event)) */



        /* Exposure */

        if(XCheckWindowEvent(dsp, win, ExposureMask, &event))
        {
                puts("   ***   E X P O S U R E   ***");
                dispatch(DISPLAY_REFRESH, 0);
        }

        

#ifdef DEBUG
        if(prefs.debug & DBG_CONTR_BUTTONS)
                printf
                (
                        "Controller: "
                        "0x%04hx %+4d %+4d (0x%08lx) (0xBBBBXXYY) "
                        "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",

                        keystate, stick_x, stick_y,
                        (WORD)(keystate << 16) | ((WORD)(HWORD)(stick_x << 8)) | ((WORD)(BYTE)stick_y),
                        (keystate & 0x8000) ? "A "      : "",
                        (keystate & 0x4000) ? "B "      : "",
                        (keystate & 0x2000) ? "Z "      : "",
                        (keystate & 0x1000) ? "START "  : "",
                        (keystate & 0x0800) ? "Dup "    : "",
                        (keystate & 0x0400) ? "Ddown "  : "",
                        (keystate & 0x0200) ? "Dleft "  : "",
                        (keystate & 0x0100) ? "Dright " : "",
                        (keystate & 0x0080) ? ""        : "",
                        (keystate & 0x0040) ? ""        : "",
                        (keystate & 0x0020) ? "L "      : "",
                        (keystate & 0x0010) ? "R "      : "",
                        (keystate & 0x0008) ? "Cup "    : "",
                        (keystate & 0x0004) ? "Cdown "  : "",
                        (keystate & 0x0002) ? "Cleft "  : "",
                        (keystate & 0x0001) ? "Cright " : "" 
                );
#endif



        return(((WORD)(keystate << 16)) | ((WORD)(HWORD)(stick_x << 8)) | ((WORD)(BYTE)stick_y));

} /* WORD CheckVisual() */










